unit d2i_support;
{User-facing I/O support routines (screen, messages, beeps, etc.)}

interface

uses
  d2i_globals;

Function DriveStatusString(status:byte):string;
Function SecsToMins(secs:real):string;
Function UpString(name:string):string;
Procedure Blip(freq,duration:word);
Procedure ChkForUserAbort;
Procedure DisplayDriveStatus(_track,_head,_sector:byte;timeonly:boolean);
Procedure DisplayReadMatrix;
Procedure DisplaySector(secbuf:pointer;line:byte);
Procedure DisplayStatus(s:string);
Procedure DoneScreen;
Procedure FlushImage;
Procedure GetOptions;
Procedure PositionCursor(x,y:byte);
Procedure PrintHeader(progname,progdesc,progvers:string;progdate:longint);
Procedure PrintHelp;
Procedure PrintWelcome;
Procedure ReconstructSector;
Procedure SetupScreen;
Procedure UpdateSalvageCounts;

var
  {
  To attempt reconstructing a bad sector, we'll read N iterations and keep
  a total of how many bits were found in each byte of the sector.  After
  enough iterations, we'll divide each count by N and round the result to
  construct a new sector from the averaged bits.
  Sector bits are from 0..(size*8)-1, with (size*8) holding N (# of iter.).
  }
  salvageData:array[0..(sectorSize*8)] of word;

  {we will read sectors directly into a global var array so that we can index
  both the damaged data and the salvage array values with an implicit DS:}
  salvagedSector:tsector;
  salvagedSectorp:psector;

implementation

uses
  cmdlin,
  support,
  streams,
  TFastWrite,
  TInterrupts;

Procedure blip(freq,duration:word);
var
  loops:word;
  l:longint;
begin
  {sound}
  Chan2SquarewaveOn(freq);
  {delay}
  l:=tickssincemidnight;
  for loops:=(duration div 18) downto 0 do begin
    repeat until l<>tickssincemidnight;
    l:=tickssincemidnight;
  end;
  {nosound}
  Chan2SquarewaveOff;
end;

Function upstring(name:string):string;
var
  foo:string;
  loop:byte;
begin
  for loop := 1 to Length(name) do
    foo[loop] := UpCase(name[loop]);
  foo[0]:=name[0];
  upstring:=foo;
end;

Procedure PrintHeader(progname,progdesc,progvers:string;progdate:longint);
const
  spacer='  ';
{prints a standard welcome header for the utilities}
begin
  Writeln(progname,spacer,progdesc,spacer,'Version ',progvers,spacer,progdate);
  {Writeln('Written by trixter@oldskool.org for the (now defunct) Flopper Project.');}
  {writeln('Use /? for available command-line options.');}
end;

Function DriveStatusString(status:byte):string;
begin
  case status of
    $00:DriveStatusString:='00: status OK';
    $01:DriveStatusString:='01: invalid INT13 parameter';
    $02:DriveStatusString:='02: address mark not found';
    $03:DriveStatusString:='03: disk write-protected';
    $04:DriveStatusString:='04: sector not found';
    $06:DriveStatusString:='06: floppy disk removed';
    $08:DriveStatusString:='08: DMA overrun';
    $09:DriveStatusString:='09: DMA past 64K boundary';
    $0C:DriveStatusString:='0C: bad media type';
    $10:DriveStatusString:='10: uncorrectable CRC error';
    $20:DriveStatusString:='20: controller failed';
    $40:DriveStatusString:='40: seek failed';
    $80:DriveStatusString:='80: timeout (disk not ready)';

    { hard-drive and special codes, for anyone interested }
    (*
    $05:DriveStatusString:='05: reset failed (hard disk)';
    $07:DriveStatusString:='07: drive parameter activity failed (hard disk)';
    $0A:DriveStatusString:='0A: bad sector detected (hard disk)';
    $0B:DriveStatusString:='0B: bad track detected (hard disk)';
    $0D:DriveStatusString:='0D: invalid number of sectors on format (PS/2 hard disk)';
    $0E:DriveStatusString:='0E: control data address mark detected (hard disk)';
    $0F:DriveStatusString:='0F: DMA arbitration level out of range (hard disk)';
    $10:DriveStatusString:='10: uncorrectable CRC or ECC data error!';
    $11:DriveStatusString:='11: data ECC corrected (hard disk)';
    $31:DriveStatusString:='31: no media in drive (IBM/MS INT 13 extensions)';
    $32:DriveStatusString:='32: incorrect drive type stored in CMOS (Compaq)';
    $AA:DriveStatusString:='AA: drive not ready (hard disk)';
    $B0:DriveStatusString:='B0: volume not locked in drive (INT 13 extensions)';
    $B1:DriveStatusString:='B1: volume locked in drive (INT 13 extensions)';
    $B2:DriveStatusString:='B2: volume not removable (INT 13 extensions)';
    $B3:DriveStatusString:='B3: volume in use (INT 13 extensions)';
    $B4:DriveStatusString:='B4: lock count exceeded (INT 13 extensions)';
    $B5:DriveStatusString:='B5: valid eject request failed (INT 13 extensions)';
    $BB:DriveStatusString:='BB: undefined error (hard disk)';
    $CC:DriveStatusString:='CC: write fault (hard disk)';
    $E0:DriveStatusString:='E0: status register error (hard disk)';
    $FF:DriveStatusString:='FF: sense operation failed (hard disk)';
    *)
  else
    DriveStatusString:='unknown error';
  end;
end;

Procedure PrintWelcome;
var
  foo:byte;
begin
  write('');
  PrintHeader('DISK2IMG','Dumps a diskette to a raw image file','2.01',20220104);
end;

function secstomins(secs:real):string;
var
  min,sec:string;
begin
  min:=inttostr(round(secs) div 60);
  sec:=inttostr(round(secs) mod 60);
  if sec[0]=#1
    then secstomins:=min+':0'+sec
    else secstomins:=min+':'+sec;
end;

Procedure DisplayStatus(s:string);
const
  blanks:string='                                                                               ';
begin
  tfastwritexy(@blanks,vidp,0,statusline,emphcolor);
  tfastwritexy(@s     ,vidp,0,statusline,emphcolor);
end;

Procedure DisplayDriveStatus(_track,_head,_sector:byte;timeonly:boolean);

var
  percent:real;
  totalsecs:real;
  secselapsed:real;
  secsremaining:real;
  temps:string;
  KBperS:word;

begin
  if not timeonly then begin
    tfastsetpos(0,progressline);
    tfastwriteln('  Drive: '+chr(65+CurDrive)+':');
    tfastwriteln('  Track: '+inttostr(_track)+' ');
    tfastwriteln('   Head: '+inttostr(_head)+' ');
    if _sector<>0
      then tfastwriteln(' Sector: '+inttostr(_sector)+'    ');
  end;

  secselapsed:=(TicksSinceMidnight-starttick)/tickHz;
  if _sector=0 {doing a track dump}
    then percent:=(trackreadcount/totaltrackreads) * 100
    else percent:=(sectorcount/sectorsremaining) * 100;
  if (percent<>0)
    then totalsecs:=((TicksSinceMidnight-starttick)/tickHZ) / percent * 100
    else totalsecs:=1;
  secsremaining:=totalsecs-secselapsed;

  tfastsetpos(0,progressline+4);

  Str(percent:2:1,temps);
  temps:=' Dumped: '+temps+'% ';
  tfastwriteln(temps);

  temps:='Done in: '+secstomins(secsremaining)+'  ';
  tfastwriteln(temps);

  if secselapsed <> 0
    then KBperS:=round( (imagesize * percent) / secselapsed / 100) div 1024
    else KBperS:=0;
  temps:='  Speed: '+inttostr(KBperS)+'KB/S ';
  tfastwriteln(temps);
end;

Procedure SetupScreen;
begin
  {re-init BIOS mode we're already in to be kind to screen accelerator TSRs}
  asm
        mov     ax,0f00h
        int     10h                     {get existing mode}
        xor     ah,ah
        int     10h                     {(re)-set existing mode}
  end;
  {Set up fast textmode writing}
  tfastinit(80,25);
  tfPreventSnow:=false;
  tfastclrscr(vidp,$0f,0);
  tfastsetcolor(normcolor);
  tfastsetdest(vidp);
  tfastsetpos(0,0);
end;

Procedure DoneScreen;
begin
  tfastclrscr(vidp,$07,0);
  PositionCursor(0,23);
end;

procedure PositionCursor(x,y:byte); assembler;
asm
        {position "cursor"}
        mov     ah,02
        mov     bh,0
        mov     dh,y
        mov     dl,x
        int     10h
end;

Procedure DisplaySector(secbuf:pointer;line:byte);assembler;
asm
        push    ds
        {mov     dx,m6845_status
        push    dx}
        xor     dx,dx
        mov     ax,(ScreenColumns * 2)
        mul     line
        les     di,vidp
        add     di,ax
        lds     si,secbuf
        add     di,(ScreenColumns * 2)-(secdisplaylinelength * 2) {location on screen to start}
        mov     ax,sectorsize
        mov     dl,secdisplaylinelength
        div     dl {ax contains sectorsize divided by sectordisplay line length which is number of lines to draw (8 usually)}
        mov     bh,al
        {mov     bl,c_vertical_sync}
        {pop     dx}
        cld

(*
I've decided to disable CGA "snow" handling for this part because the delay
can take just enough time that it causes an extra disk revolution.
An extra revolution lasts *much* longer than "snow" is visible on-screen.
@WDR: {wait during retrace, because we don't know where we are in the cycle}
        in      al,dx
        test    al,bl {if our bit is 1, then we're already in retrace, which means we missed it}
        jnz     @WDR  {jump if 1 (not 0) = keep looping as long as we're retracing}
@WDD: {wait for display to be over}
        in      al,dx
        test    al,bl
        jz      @WDD  {loop until we aren't drawing any more (ie. retracing)}
*)

@displayloop:
        mov     cx,(secdisplaylinelength / 16) {unrolled}
@lineloop: {unrolled to 16 items}
        (*
        lodsb;stosw > movsb;inc di in terms of speed
        *)
        movsb; inc di; movsb; inc di; movsb; inc di; movsb; inc di;
        movsb; inc di; movsb; inc di; movsb; inc di; movsb; inc di;
        movsb; inc di; movsb; inc di; movsb; inc di; movsb; inc di;
        movsb; inc di; movsb; inc di; movsb; inc di; movsb; inc di;
        loop    @lineloop
        add     di,(ScreenColumns * 2)-(secdisplaylinelength * 2) {skip to next line}
        dec     bh                      {dec line count}
        jnz     @displayloop            {if not done, do another line}
        pop     ds
end;

Procedure updateSalvageCounts;assembler;
{updates bit counter statistics}
asm
        mov     si,offset salvagedSector
        mov     di,offset salvageData
        mov     cx,sectorsize
        xor     bx,bx
        cld
@countbits:
        lodsb
        shl     al,1
        adc     word ptr [di],bx
        shl     al,1
        adc     word ptr [di+2],bx
        shl     al,1
        adc     word ptr [di+4],bx
        shl     al,1
        adc     word ptr [di+6],bx
        shl     al,1
        adc     word ptr [di+8],bx
        shl     al,1
        adc     word ptr [di+10],bx
        shl     al,1
        adc     word ptr [di+12],bx
        shl     al,1
        adc     word ptr [di+14],bx
        add     di,16
        loop    @countbits
end;

Procedure reconstructSector; assembler;
{
The idea is to write the sector out using the average of each read bit.
For each original bit position in the sector, the total bits read is
divided by the # of reads, rounded up (meaning, 0.5 or higher is 1,
otherwise 0).  salvagedSector and salvageData are both in DS.
salvageData[(512*8)*2] holds the # of reads.
}
asm
@start:
        push    bp
        mov     si,offset salvageData
        mov     di,offset salvagedSector
        mov     bp,word ptr salvageData[(sectorsize*8)*2] {ax:=count / total}
        mov     cx,sectorsize
        push    ds
        pop     es
        cld
@bytesloop:
        xor     bx,bx                   {reconstructed byte}

        {unrolled 8 times, once for each bit, so we can gain a register}
        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        xchg    bx,ax
        stosb                           {store byte to sector buffer}

        loop    @bytesloop              {do all bytes in the sector}
        pop     bp
end;

Procedure printhelp;
begin
  writeln('Usage: "DISK2IMG d: filename.ext [options]", where "d:" is drive letter.');
  writeln;
  writeln('Option  Description                                           Default');
  writeln('  /Txx  number of Tracks                                      /t',numtracks);
  writeln('  /Hx   number of Heads                                       /h',numheads);
  writeln('  /Sxx  number of Sectors per track                           /s',numsectors);
  writeln('  /Axxx Number of sector read attempts                        ',maxRecoverCount);
  writeln('  /Q    Query boot sector to get disk parameters              ',Query);
  writeln('  /E    Rename file extension to match image size             ',renameExt);
  writeln('  /I    Ignore errors (only useful for non-DOS disks)         ',ignoreErrors);
  writeln('  /R    Reconstruct bad sectors (CAUTION! SEE DOCUMENTATION)  ',Reconstruct);
  writeln;
  writeln('Default settings reflect a ',numtracks*numsectors,'K diskette in drive ',upcase(chr(65+CurDrive)),':');
  writeln('If no filename is specified, the name "'+defaultname+'" will be used.');
  writeln;
  writeln('Examples:');
  writeln('DISK2IMG a: mygame.img                         dumps disk in A: to "mygame.img"');
  writeln('DISK2IMG b: bootdisk.img /q /e    query disk geometry and rewrite .img to match');
  halt(1);
end;

Procedure GetOptions;
var
  foos:string;
  fooc:char;

begin
  if is_param('?') or (non_flag_count=0) then begin
    writeln(#13#10'Error: Minimum command-line arguments must provide either drive or filename.');
    printhelp;
  end;

  {new filename?}
  imageName:=defaultName;
  if non_flag_count>0 then begin
    imageName:=non_flag_param(1);
    {if filename is "A:", "B:" etc. then pick the other string as filename}
    if imageName[2]=':'
      then if non_flag_count>1
        then imageName:=non_flag_param(2)
        else imageName:=defaultName;
  end;

  {Also handle if first string is less than 3 chars since that is a drive}
  if (non_flag_count<>0) and (length(non_flag_param(1))<3) then begin
    foos:=non_flag_param(1);
    CurDrive:=ord(upcase(foos[1]))-65;
  end;

  {number of tracks}
  if is_param('t') then numtracks:=param_int('t');

  {number of sectors per track}
  if is_param('s') then numsectors:=param_int('s');

  {number of heads}
  if is_param('h') then numheads:=param_int('h');

  {number of recovery attempts}
  if is_param('a') then maxRecoverCount:=param_int('a');
  if maxRecoverCount>(60*300)
    then fatalerror(99,inttostr(maxRecoverCount)+' attempts takes '
         +inttostr(maxRecoverCount div 300)+' minutes per sector and will shred your disk!');

  {query disk boot record for track, heads, spt}
  if is_param('q') then Query:=true;

  {Rename filename extension to match size of image?}
  if is_param('e') then renameExt:=true;

  {Ignore drive reset errors -- good for trying to recover bad disks}
  if is_param('i') then ignoreErrors:=true;

  {Attempt to salvage bad sectors by reconstructing them}
  if is_param('r') then Reconstruct:=true;
end;

Procedure FlushImage;
begin
  DisplayStatus('Flushing temporary data into image file...');

  imageTemp^.seek(0);
  imageFile^.seek(0);
  FastCopy(imageTemp^,imageFile^,imagesize);

  DisplayStatus('Closing image file...');
  dispose(imageFile,done);

  DisplayStatus('Disposing RAM image...');
  dispose(imageTemp,done);
end;

Procedure chkForUserAbort;
var
  ch:char;
begin
  if keypressed then begin
    ch:=readkeychar;
    if ch=#27 then begin
      FlushImage;
      fatalerror(0,'User aborted');
    end;
  end;
end;

Procedure DisplayReadMatrix;
{number of sectors per track = # of lines we need onscreen}
var
  curline:byte;
  _track,_head,_sector:byte;
  c:char;
  b:byte;
  sofs:word;
begin
  if totalsectors <= 720 then begin
    {40-track disks, we can display every sector as it appears in the image}
    curline:=statusline-numSectors;
    sofs:=curline*160;
    for _track:=0 to numtracks-1 do
      for _head:=0 to numheads-1 do
        for _sector:=1 to numsectors do begin
          c:=readMatrixIcons[ord(readStatus[_track,_head,_sector])];
          asm
            les di,vidp
            add di,sofs
            mov ah,sectcolor
            mov al,c
            stosw
            mov sofs,di
          end;
        end;
  end else if totalsectors<=1440 then begin
    {720K disks, we can use same onscreen char for both heads}
    {How to combine heads?  "good"=0 so we'll just XOR them together}
    {We are assuming there are no single-sided 3.5" disks out there.}
    curline:=statusline-numSectors;
    sofs:=curline*160;
    for _track:=0 to numtracks-1 do
      for _sector:=1 to numsectors do begin
        c := readMatrixIcons[(byte(readStatus[_track,0,_sector])
                           OR byte(readStatus[_track,1,_sector]))];
        asm
          les di,vidp
          add di,sofs
          mov ah,sectcolor
          mov al,c
          stosw
          mov sofs,di
        end;
      end;
  end else begin
    {1.2 and 1.44 MB disks: display track+heads only}
    curline:=statusline-numheads;
    sofs:=curline*160;
    for _head:=0 to numheads-1 do begin
      b:=0;
      for _track:=0 to numtracks-1 do begin
        for _sector:=1 to numsectors do
          b:=b OR byte(readStatus[_track,0,_sector]);
        c := readMatrixIcons[b];
        asm
          les di,vidp
          add di,sofs
          mov ah,sectcolor
          mov al,c
          stosw
          mov sofs,di
        end;
      end;
    end;
  end;
end;

end.
